#===============================================================================
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import pytest
import numpy as np
from numpy.testing import assert_array_equal, assert_allclose, assert_array_almost_equal
from sklearn import datasets
from sklearn.metrics.pairwise import rbf_kernel

from onedal.svm import NuSVR
from sklearn.svm import NuSVR as SklearnNuSVR

from onedal.tests.utils._device_selection import (get_queues,
                                                  pass_if_not_implemented_for_gpu)


@pass_if_not_implemented_for_gpu(reason="nusvr is not implemented")
@pytest.mark.parametrize('queue', get_queues())
def test_diabetes_simple(queue):
    diabetes = datasets.load_diabetes()
    clf = NuSVR(kernel='linear', C=10.)
    clf.fit(diabetes.data, diabetes.target, queue=queue)
    assert clf.score(diabetes.data, diabetes.target, queue=queue) > 0.02


@pass_if_not_implemented_for_gpu(reason="nusvr is not implemented")
@pytest.mark.parametrize('queue', get_queues())
def test_input_format_for_diabetes(queue):
    diabetes = datasets.load_diabetes()

    c_contiguous_numpy = np.asanyarray(diabetes.data, dtype='float', order='C')
    assert c_contiguous_numpy.flags.c_contiguous
    assert not c_contiguous_numpy.flags.f_contiguous
    assert not c_contiguous_numpy.flags.fnc

    clf = NuSVR(kernel='linear', C=10.)
    clf.fit(c_contiguous_numpy, diabetes.target, queue=queue)
    dual_c_contiguous_numpy = clf.dual_coef_
    res_c_contiguous_numpy = clf.predict(c_contiguous_numpy, queue=queue)

    f_contiguous_numpy = np.asanyarray(diabetes.data, dtype='float', order='F')
    assert not f_contiguous_numpy.flags.c_contiguous
    assert f_contiguous_numpy.flags.f_contiguous
    assert f_contiguous_numpy.flags.fnc

    clf = NuSVR(kernel='linear', C=10.)
    clf.fit(f_contiguous_numpy, diabetes.target, queue=queue)
    dual_f_contiguous_numpy = clf.dual_coef_
    res_f_contiguous_numpy = clf.predict(f_contiguous_numpy, queue=queue)
    assert_allclose(dual_c_contiguous_numpy, dual_f_contiguous_numpy)
    assert_allclose(res_c_contiguous_numpy, res_f_contiguous_numpy)


@pass_if_not_implemented_for_gpu(reason="nusvr is not implemented")
@pytest.mark.parametrize('queue', get_queues())
def test_predict(queue):
    iris = datasets.load_iris()
    X = iris.data
    y = iris.target

    reg = NuSVR(kernel='linear', C=0.1).fit(X, y, queue=queue)

    linear = np.dot(X, reg.support_vectors_.T)
    dec = np.dot(linear, reg.dual_coef_.T) + reg.intercept_
    assert_array_almost_equal(dec.ravel(), reg.predict(X, queue=queue).ravel())

    reg = NuSVR(kernel='rbf', gamma=1).fit(X, y, queue=queue)

    rbfs = rbf_kernel(X, reg.support_vectors_, gamma=reg.gamma)
    dec = np.dot(rbfs, reg.dual_coef_.T) + reg.intercept_
    assert_array_almost_equal(dec.ravel(), reg.predict(X, queue=queue).ravel())


def _test_diabetes_compare_with_sklearn(queue, kernel):
    diabetes = datasets.load_diabetes()
    clf_onedal = NuSVR(kernel=kernel, nu=.25, C=10.)
    clf_onedal.fit(diabetes.data, diabetes.target, queue=queue)
    result = clf_onedal.score(diabetes.data, diabetes.target, queue=queue)

    clf_sklearn = SklearnNuSVR(kernel=kernel, nu=.25, C=10.)
    clf_sklearn.fit(diabetes.data, diabetes.target)
    expected = clf_sklearn.score(diabetes.data, diabetes.target)

    assert result > expected - 1e-5
    assert_allclose(clf_sklearn.intercept_, clf_onedal.intercept_, atol=1e-3)
    assert_allclose(clf_sklearn.support_vectors_.shape,
                    clf_sklearn.support_vectors_.shape)
    assert_allclose(clf_sklearn.dual_coef_, clf_onedal.dual_coef_, atol=1e-2)


@pass_if_not_implemented_for_gpu(reason="nusvr is not implemented")
@pytest.mark.parametrize('queue', get_queues())
@pytest.mark.parametrize('kernel', ['linear', 'rbf', 'poly', 'sigmoid'])
def test_diabetes_compare_with_sklearn(queue, kernel):
    _test_diabetes_compare_with_sklearn(queue, kernel)


def _test_boston_rbf_compare_with_sklearn(queue, C, nu, gamma):
    diabetes = datasets.load_boston()

    clf = NuSVR(kernel='rbf', gamma=gamma, C=C, nu=nu)
    clf.fit(diabetes.data, diabetes.target, queue=queue)
    result = clf.score(diabetes.data, diabetes.target, queue=queue)

    clf = SklearnNuSVR(kernel='rbf', gamma=gamma, C=C, nu=nu)
    clf.fit(diabetes.data, diabetes.target)
    expected = clf.score(diabetes.data, diabetes.target)

    assert result > 0.4
    assert abs(result - expected) < 1e-3


@pass_if_not_implemented_for_gpu(reason="nusvr is not implemented")
@pytest.mark.parametrize('queue', get_queues())
@pytest.mark.parametrize('gamma', ['scale', 'auto'])
@pytest.mark.parametrize('C', [100.0, 1000.0])
@pytest.mark.parametrize('nu', [0.25, 0.75])
def test_boston_rbf_compare_with_sklearn(queue, C, nu, gamma):
    _test_boston_rbf_compare_with_sklearn(queue, C, nu, gamma)


def _test_boston_linear_compare_with_sklearn(queue, C, nu):
    diabetes = datasets.load_boston()

    clf = NuSVR(kernel='linear', C=C, nu=nu)
    clf.fit(diabetes.data, diabetes.target, queue=queue)
    result = clf.score(diabetes.data, diabetes.target, queue=queue)

    clf = SklearnNuSVR(kernel='linear', C=C, nu=nu)
    clf.fit(diabetes.data, diabetes.target)
    expected = clf.score(diabetes.data, diabetes.target)

    assert result > 0.5
    assert abs(result - expected) < 1e-3


@pass_if_not_implemented_for_gpu(reason="nusvr is not implemented")
@pytest.mark.parametrize('queue', get_queues())
@pytest.mark.parametrize('C', [0.001, 0.1])
@pytest.mark.parametrize('nu', [0.25, 0.75])
def test_boston_linear_compare_with_sklearn(queue, C, nu):
    _test_boston_linear_compare_with_sklearn(queue, C, nu)


def _test_boston_poly_compare_with_sklearn(queue, params):
    diabetes = datasets.load_boston()

    clf = NuSVR(kernel='poly', **params)
    clf.fit(diabetes.data, diabetes.target, queue=queue)
    result = clf.score(diabetes.data, diabetes.target, queue=queue)

    clf = SklearnNuSVR(kernel='poly', **params)
    clf.fit(diabetes.data, diabetes.target)
    expected = clf.score(diabetes.data, diabetes.target)

    assert result > 0.5
    assert abs(result - expected) < 1e-3


@pass_if_not_implemented_for_gpu(reason="nusvr is not implemented")
@pytest.mark.parametrize('queue', get_queues())
@pytest.mark.parametrize('params', [
    {'degree': 2, 'coef0': 0.1, 'gamma': 'scale', 'C': 100, 'nu': .25},
    {'degree': 3, 'coef0': 0.0, 'gamma': 'scale', 'C': 1000, 'nu': .75}
])
def test_boston_poly_compare_with_sklearn(queue, params):
    _test_boston_poly_compare_with_sklearn(queue, params)


@pass_if_not_implemented_for_gpu(reason="nusvr is not implemented")
@pytest.mark.parametrize('queue', get_queues())
def test_pickle(queue):
    diabetes = datasets.load_diabetes()

    clf = NuSVR(kernel='rbf', C=10.)
    clf.fit(diabetes.data, diabetes.target, queue=queue)
    expected = clf.predict(diabetes.data, queue=queue)

    import pickle
    dump = pickle.dumps(clf)
    clf2 = pickle.loads(dump)

    assert type(clf2) == clf.__class__
    result = clf2.predict(diabetes.data, queue=queue)
    assert_array_equal(expected, result)



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# daal4py Ridge Regression example for streaming on shared memory systems

import daal4py as d4p
import numpy as np

# let's try to use pandas' fast csv reader
try:
    import pandas

    def read_csv(f, c, s=0, n=None, t=np.float64):
        return pandas.read_csv(f, usecols=c, delimiter=',', header=None,
                               skiprows=s, nrows=n, dtype=t)
except:
    # fall back to numpy genfromtxt
    def read_csv(f, c, s=0, n=np.iinfo(np.int64).max):
        a = np.genfromtxt(f, usecols=c, delimiter=',', skip_header=s, max_rows=n)
        if a.shape[0] == 0:
            raise Exception("done")
        if a.ndim == 1:
            return a[:, np.newaxis]
        return a


def main(readcsv=read_csv, method='defaultDense'):
    infile = "./data/batch/linear_regression_train.csv"
    testfile = "./data/batch/linear_regression_test.csv"

    # Configure a Ridge regression training object for streaming
    train_algo = d4p.ridge_regression_training(interceptFlag=True, streaming=True)

    chunk_size = 250
    lines_read = 0
    # read and feed chunk by chunk
    while True:
        # Read data in chunks
        # Let's have 10 independent, and 2 dependent variables (for each observation)
        try:
            indep_data = readcsv(infile, range(10), lines_read, chunk_size)
            dep_data = readcsv(infile, range(10, 12), lines_read, chunk_size)
        except:
            break
        # Now feed chunk
        train_algo.compute(indep_data, dep_data)
        lines_read += indep_data.shape[0]

    # All chunks are done, now finalize the computation
    train_result = train_algo.finalize()

    # Now let's do some prediction
    predict_algo = d4p.ridge_regression_prediction()
    # read test data (with same #features)
    pdata = readcsv(testfile, range(10))
    ptdata = readcsv(testfile, range(10, 12))
    # now predict using the model from the training above
    predict_result = predict_algo.compute(pdata, train_result.model)

    # The prediction result provides prediction
    assert predict_result.prediction.shape == (pdata.shape[0], dep_data.shape[1])

    return (predict_result, ptdata)


if __name__ == "__main__":
    (predict_result, ptdata) = main()
    print(
        "\nRidge Regression prediction results: (first 10 rows):\n",
        predict_result.prediction[0:10]
    )
    print("\nGround truth (first 10 rows):\n", ptdata[0:10])
    print('All looks good!')



#===============================================================================
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import pytest
import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal

from onedal.svm import SVC

from sklearn.utils.estimator_checks import check_estimator
import sklearn.utils.estimator_checks
from sklearn import datasets
from sklearn.metrics.pairwise import rbf_kernel
from sklearn.datasets import make_blobs
from sklearn.model_selection import train_test_split

from onedal.tests.utils._device_selection import (get_queues,
                                                  pass_if_not_implemented_for_gpu)


def _replace_and_save(md, fns, replacing_fn):
    saved = dict()
    for check_f in fns:
        try:
            fn = getattr(md, check_f)
            setattr(md, check_f, replacing_fn)
            saved[check_f] = fn
        except RuntimeError:
            pass
    return saved


def _restore_from_saved(md, saved_dict):
    for check_f in saved_dict:
        setattr(md, check_f, saved_dict[check_f])


def test_estimator():
    def dummy(*args, **kwargs):
        pass

    md = sklearn.utils.estimator_checks
    saved = _replace_and_save(md, [
        'check_sample_weights_invariance',  # Max absolute difference: 0.0008
        'check_estimators_fit_returns_self',  # ValueError: empty metadata
        'check_classifiers_train',  # assert y_pred.shape == (n_samples,)
        'check_estimators_unfitted',  # Call 'fit' with appropriate arguments
        'check_estimator_sparse_data',  # This fails with scikit-learn v1.0.2
    ], dummy)
    check_estimator(SVC())
    _restore_from_saved(md, saved)


def _test_libsvm_parameters(queue, array_constr, dtype):
    X = array_constr([[-2, -1], [-1, -1], [-1, -2],
                      [1, 1], [1, 2], [2, 1]], dtype=dtype)
    y = array_constr([1, 1, 1, 2, 2, 2], dtype=dtype)

    clf = SVC(kernel='linear').fit(X, y, queue=queue)
    assert_array_equal(clf.dual_coef_, [[-0.25, .25]])
    assert_array_equal(clf.support_, [1, 3])
    assert_array_equal(clf.support_vectors_, (X[1], X[3]))
    assert_array_equal(clf.intercept_, [0.])
    assert_array_equal(clf.predict(X), y)


# TODO: investigate sporadic failures on GPU
@pytest.mark.parametrize('queue', get_queues('host,cpu'))
@pytest.mark.parametrize('array_constr', [np.array])
@pytest.mark.parametrize('dtype', [np.float32, np.float64])
def test_libsvm_parameters(queue, array_constr, dtype):
    _test_libsvm_parameters(queue, array_constr, dtype)


@pytest.mark.parametrize('queue', get_queues('cpu') + [
    pytest.param(get_queues('gpu'),
                 marks=pytest.mark.xfail(
                     reason="class weights are not implemented "
                            "but the error is not raised"))])
def test_class_weight(queue):
    X = np.array([[-2, -1], [-1, -1], [-1, -2], [1, 1], [1, 2], [2, 1]])
    y = np.array([1, 1, 1, 2, 2, 2])

    clf = SVC(class_weight={1: 0.1})
    clf.fit(X, y, queue=queue)
    assert_array_almost_equal(clf.predict(X, queue=queue), [2] * 6)


# TODO: investigate sporadic failures on GPU
@pytest.mark.parametrize('queue', get_queues('host,cpu'))
def test_sample_weight(queue):
    X = np.array([[-2, 0], [-1, -1], [0, -2], [0, 2], [1, 1], [2, 2]])
    y = np.array([1, 1, 1, 2, 2, 2])

    clf = SVC(kernel='linear')
    clf.fit(X, y, sample_weight=[1] * 6, queue=queue)
    assert_array_almost_equal(clf.intercept_, [0.0])


@pytest.mark.parametrize('queue', get_queues())
def test_decision_function(queue):
    X = np.array([[-2, -1], [-1, -1], [-1, -2], [1, 1], [1, 2], [2, 1]], dtype=np.float32)
    Y = np.array([1, 1, 1, 2, 2, 2], dtype=np.float32)

    clf = SVC(kernel='rbf', gamma=1, decision_function_shape='ovo')
    clf.fit(X, Y, queue=queue)

    rbfs = rbf_kernel(X, clf.support_vectors_, gamma=clf.gamma)
    dec = np.dot(rbfs, clf.dual_coef_.T) + clf.intercept_
    assert_array_almost_equal(dec.ravel(), clf.decision_function(X, queue=queue))


@pass_if_not_implemented_for_gpu(reason="multiclass svm is not implemented")
@pytest.mark.parametrize('queue', get_queues())
def test_iris(queue):
    iris = datasets.load_iris()
    clf = SVC(kernel='linear').fit(iris.data, iris.target, queue=queue)
    assert clf.score(iris.data, iris.target, queue=queue) > 0.9
    assert_array_equal(clf.classes_, np.sort(clf.classes_))


@pass_if_not_implemented_for_gpu(reason="multiclass svm is not implemented")
@pytest.mark.parametrize('queue', get_queues())
def test_decision_function_shape(queue):
    X, y = make_blobs(n_samples=80, centers=5, random_state=0)
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

    # check shape of ovo_decition_function=True
    clf = SVC(kernel='linear',
              decision_function_shape='ovo').fit(X_train, y_train, queue=queue)
    dec = clf.decision_function(X_train, queue=queue)
    assert dec.shape == (len(X_train), 10)

    with pytest.raises(ValueError, match="must be either 'ovr' or 'ovo'"):
        SVC(decision_function_shape='bad').fit(X_train, y_train, queue=queue)


@pass_if_not_implemented_for_gpu(reason="multiclass svm is not implemented")
@pytest.mark.parametrize('queue', get_queues())
def test_pickle(queue):
    iris = datasets.load_iris()
    clf = SVC(kernel='linear').fit(iris.data, iris.target, queue=queue)
    expected = clf.decision_function(iris.data, queue=queue)

    import pickle
    dump = pickle.dumps(clf)
    clf2 = pickle.loads(dump)

    assert type(clf2) == clf.__class__
    result = clf2.decision_function(iris.data, queue=queue)
    assert_array_equal(expected, result)


@pass_if_not_implemented_for_gpu(reason="sigmoid kernel is not implemented")
@pytest.mark.parametrize('queue', get_queues('cpu') + [
    pytest.param(get_queues('gpu'),
                 marks=pytest.mark.xfail(reason="raises Unimplemented error "
                                                "with inconsistent error message"))])
@pytest.mark.parametrize('dtype', [np.float32, np.float64])
def test_svc_sigmoid(queue, dtype):
    X_train = np.array([[-1, 2], [0, 0], [2, -1],
                        [+1, +1], [+1, +2], [+2, +1]], dtype=dtype)
    X_test = np.array([[0, 2], [0.5, 0.5],
                       [0.3, 0.1], [2, 0], [-1, -1]], dtype=dtype)
    y_train = np.array([1, 1, 1, 2, 2, 2], dtype=dtype)
    svc = SVC(kernel='sigmoid').fit(X_train, y_train, queue=queue)

    assert_array_equal(svc.dual_coef_, [[-1, -1, -1, 1, 1, 1]])
    assert_array_equal(svc.support_, [0, 1, 2, 3, 4, 5])
    assert_array_equal(svc.predict(X_test, queue=queue), [2, 2, 1, 2, 1])



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# daal4py DecisionTree scikit-learn-compatible estimator classes

import numpy as np
import numbers
import warnings
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.exceptions import DataConversionWarning
from sklearn.utils.validation import (
    check_array, check_is_fitted, check_consistent_length
)
from sklearn.utils.multiclass import check_classification_targets
import daal4py as d4p
from .._utils import (make2d, getFPType)
from .._device_offload import support_usm_ndarray
from scipy.sparse import issparse


class DecisionTreeClassifier(BaseEstimator, ClassifierMixin):
    """
    Decision tree classifier powered by Intel(R) oneAPI Data Analytics Library.

       https://software.intel.com/en-us/daal-programming-guide-decision-tree-2
       https://software.intel.com/en-us/daal-programming-guide-batch-processing-50

    Parameters
    ----------
    max_depth : int or None, default=None
        Depth of the tree fitten to data. None corresponds to unlimited depth.

    min_observations_in_leaf_node : int, default=10
        The number of estimators in the ensemble.

    estimator_params : list of str, default=tuple()
        The list of attributes to use as parameters when instantiating a
        new base estimator. If none are given, default parameters are used.

    Attributes
    ----------
    base_estimator_ : estimator
        The base estimator from which the ensemble is grown.

    estimators_ : list of estimators
        The collection of fitted base estimators.


    Training:
        inputs: dataForPruning, labelsForPruning
        parameters: fptype, method, nClasses, splitCriterion, pruning,
                    maxTreeDepth, minObservationsInLeafNodes

    Prediction:
        parameters: fptype, method, nBins, nClasses,
        resultsToEvaluate (computeClassesLabels|computeClassProbabilities)
        N.B.: The only supported value for current version of the library is nBins=1.
              nBins is the number of bins used to compute probabilities of the
              observations belonging to the class.
    """
    def __init__(self, max_depth=None, min_observations_in_leaf_node=1,
                 split_criterion='gini'):
        self.max_depth = max_depth
        self.min_observations_in_leaf_node = min_observations_in_leaf_node
        self.split_criterion = split_criterion

    def _daal4py_fit(self, X, y, w, pruning_set=None):
        X_fptype = getFPType(X)
        X = make2d(X)
        y = make2d(y)

        if pruning_set is None:
            _pruning = "none"
            _pruning_X = None
            _pruning_y = None
        else:
            _pruning = "reducedErrorPruning"
            if isinstance(pruning_set, (tuple, list)) and len(pruning_set) == 2:
                _pruning_X, _pruning_y = pruning_set
                check_consistent_length(_pruning_X, _pruning_y)
                _pruning_X = make2d(_pruning_X)
                _pruning_y = make2d(_pruning_y)
            else:
                raise ValueError("pruning_set parameter is expected to be "
                                 "a tuple of pruning features and pruning "
                                 "dependent variables")

        if w is not None:
            w = make2d(np.asarray(w))

        daal_max_tree_depth = 0 if (self.max_depth is None) else int(self.max_depth) + 1
        alg = d4p.decision_tree_classification_training(
            fptype=X_fptype,
            method="defaultDense",
            nClasses=int(self.n_classes_),
            splitCriterion=self.split_criterion,
            maxTreeDepth=daal_max_tree_depth,
            minObservationsInLeafNodes=int(self.min_observations_in_leaf_node),
            pruning=_pruning)
        res = alg.compute(X, y,
                          dataForPruning=_pruning_X,
                          labelsForPruning=_pruning_y,
                          weights=w)
        self.daal_model_ = res.model
        self._cached_tree_state_ = None

    def _get_tree_state(self):
        """
        Internal utility that returns an array behind scikit-learn's tree object
        from daal_model_ produced by call to fit
        """
        check_is_fitted(self, ['daal_model_', '_cached_tree_state_'])
        if self._cached_tree_state_ is None:
            tree_state_class = d4p.getTreeState(self.daal_model_, int(self.n_classes_))
            self._cached_tree_state_ = tree_state_class
        return self._cached_tree_state_

    def get_n_leaves(self):
        ts = self._get_tree_state()
        return ts.leaf_count

    def get_depth(self):
        ts = self._get_tree_state()
        return ts.max_depth

    @support_usm_ndarray()
    def fit(self, X, y, sample_weight=None, pruning_set=None):
        """Build a decision tree classifier from the training set (X, y).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Internally, it will be converted to
            ``dtype=np.float64`` and if a sparse matrix is provided
            to a sparse ``csc_matrix``.

        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels) as integers or strings.

        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, then samples are equally weighted. Splits
            that would create child nodes with net zero or negative weight are
            ignored while searching for a split in each node. Splits are also
            ignored if they would result in any single class carrying a
            negative weight in either child node.

        pruning_set: None or a tuple of (X, y) corrsponding to features and
            associated labels used for tree pruning. See [1] for more details.

        Returns
        -------
        self : DecisionTreeClassifier
            Fitted estimator.


        [1] https://software.intel.com/en-us/daal-programming-guide-decision-tree-2
        """

        if self.split_criterion not in ('gini', 'infoGain'):
            raise ValueError('Parameter "split_criterion" must be '
                             '"gini" or "infoGain".')

        if not isinstance(self.max_depth, numbers.Integral) or \
                self.max_depth < 0:
            if self.max_depth is not None:
                raise ValueError('Parameter "max_depth" must be '
                                 'a non-negative integer value or None.')

        if not isinstance(self.min_observations_in_leaf_node, numbers.Integral) or \
                self.min_observations_in_leaf_node <= 0:
            raise ValueError('Parameter "min_observations_in_leaf_node" must be '
                             'non-zero positive integer value.')

        X = check_array(X, dtype=[np.single, np.double])
        y = np.asarray(y)
        y = np.atleast_1d(y)

        if y.ndim == 2 and y.shape[1] == 1:
            warnings.warn(
                "A column-vector y was passed when a 1d array was"
                " expected. Please change the shape of y to "
                "(n_samples,), for example using ravel().",
                DataConversionWarning, stacklevel=2
            )

        check_consistent_length(X, y)

        if y.ndim == 1:
            # reshape is necessary to preserve the data contiguity against vs
            # [:, np.newaxis] that does not.
            y = np.reshape(y, (-1, 1))

        self.n_outputs_ = y.shape[1]
        if self.n_outputs_ != 1:
            _class_name = self.__class__.__name__
            raise ValueError(_class_name + " does not currently support "
                             "multi-output data. "
                             "Consider using OneHotEncoder")

        y = check_array(y, ensure_2d=False, dtype=None)
        check_classification_targets(y)

        y = np.copy(y)

        self.classes_ = []
        self.n_classes_ = []

        y_store_unique_indices = np.zeros(y.shape, dtype=np.int)
        for k in range(self.n_outputs_):
            classes_k, y_store_unique_indices[:, k] = \
                np.unique(y[:, k], return_inverse=True)
            self.classes_.append(classes_k)
            self.n_classes_.append(classes_k.shape[0])
        y = y_store_unique_indices

        self.n_classes_ = self.n_classes_[0]
        self.classes_ = self.classes_[0]

        self.n_features_ = X.shape[1]

        if self.n_classes_ < 2:
            raise ValueError("Training data only contain information about one class.")

        self._daal4py_fit(X, y, sample_weight, pruning_set=pruning_set)

        return self

    def _validate_X_predict(self, X, check_input):
        """Validate X whenever one tries to predict, apply, predict_proba"""
        if check_input:
            X = check_array(X, dtype=[np.single, np.double], accept_sparse="csr")
            if issparse(X) and \
                    (X.indices.dtype != np.intc or X.indptr.dtype != np.intc):
                raise ValueError("No support for np.int64 index based "
                                 "sparse matrices")

        n_features = X.shape[1]
        if self.n_features_ != n_features:
            raise ValueError("Number of features of the model must "
                             "match the input. Model n_features is %s and "
                             "input n_features is %s "
                             % (self.n_features_, n_features))

        return X

    def _daal4py_predict(self, X):
        fptype = getFPType(X)
        alg = d4p.decision_tree_classification_prediction(
            fptype=fptype,
            method="defaultDense",
            nBins=1,
            nClasses=self.n_classes_,
            resultsToEvaluate="computeClassLabels"
        )
        res = alg.compute(X, self.daal_model_)
        return res.prediction.ravel()

    @support_usm_ndarray()
    def predict(self, X, check_input=True):
        check_is_fitted(self, 'daal_model_')
        X = self._validate_X_predict(X, check_input)
        y = self._daal4py_predict(X)
        return self.classes_.take(np.asarray(y, dtype=np.intp), axis=0)

    @support_usm_ndarray()
    def predict_proba(self, X, check_input=True):
        check_is_fitted(self, 'daal_model_')
        X = self._validate_X_predict(X, check_input)
        y = self._daal4py_predict(X)
        return self.classes_.take(np.asarray(y, dtype=np.intp), axis=0)



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np
import numbers
from scipy import sparse as sp

from sklearn.utils import check_random_state, check_array
from sklearn.utils.sparsefuncs import mean_variance_axis
from sklearn.utils.validation import (
    check_is_fitted,
    _num_samples,
    _deprecate_positional_args)

from sklearn.cluster._kmeans import _labels_inertia
from sklearn.utils._openmp_helpers import _openmp_effective_n_threads

from sklearn.exceptions import ConvergenceWarning
from sklearn.utils.extmath import row_norms
import warnings

from sklearn.cluster import KMeans as KMeans_original

import daal4py
from .._utils import (
    getFPType,
    sklearn_check_version,
    PatchingConditionsChain)
from .._device_offload import support_usm_ndarray


def _validate_center_shape(X, n_centers, centers):
    """Check if centers is compatible with X and n_centers"""
    if centers.shape[0] != n_centers:
        raise ValueError(
            f"The shape of the initial centers {centers.shape} does not "
            f"match the number of clusters {n_centers}.")
    if centers.shape[1] != X.shape[1]:
        raise ValueError(
            f"The shape of the initial centers {centers.shape} does not "
            f"match the number of features of the data {X.shape[1]}.")


def _tolerance(X, rtol):
    """Compute absolute tolerance from the relative tolerance"""
    if rtol == 0.0:
        return rtol
    if sp.issparse(X):
        variances = mean_variance_axis(X, axis=0)[1]
        mean_var = np.mean(variances)
    else:
        mean_var = np.var(X, axis=0).mean()
    return mean_var * rtol


def _daal4py_compute_starting_centroids(
    X,
    X_fptype,
    nClusters,
    cluster_centers_0,
    verbose,
    random_state
):
    def is_string(s, target_str):
        return isinstance(s, str) and s == target_str
    is_sparse = sp.isspmatrix(X)

    deterministic = False
    if is_string(cluster_centers_0, 'k-means++'):
        _seed = random_state.randint(np.iinfo('i').max)
        plus_plus_method = "plusPlusCSR" if is_sparse else "plusPlusDense"
        daal_engine = daal4py.engines_mt19937(
            fptype=X_fptype, method="defaultDense", seed=_seed)
        _n_local_trials = 2 + int(np.log(nClusters))
        kmeans_init = daal4py.kmeans_init(
            nClusters,
            fptype=X_fptype,
            nTrials=_n_local_trials,
            method=plus_plus_method,
            engine=daal_engine,
        )
        kmeans_init_res = kmeans_init.compute(X)
        centroids_ = kmeans_init_res.centroids
    elif is_string(cluster_centers_0, 'random'):
        _seed = random_state.randint(np.iinfo('i').max)
        random_method = "randomCSR" if is_sparse else "randomDense"
        daal_engine = daal4py.engines_mt19937(
            seed=_seed, fptype=X_fptype, method="defaultDense")
        kmeans_init = daal4py.kmeans_init(
            nClusters,
            fptype=X_fptype,
            method=random_method,
            engine=daal_engine,
        )
        kmeans_init_res = kmeans_init.compute(X)
        centroids_ = kmeans_init_res.centroids
    elif hasattr(cluster_centers_0, '__array__'):
        deterministic = True
        cc_arr = np.ascontiguousarray(cluster_centers_0, dtype=X.dtype)
        _validate_center_shape(X, nClusters, cc_arr)
        centroids_ = cc_arr
    elif callable(cluster_centers_0):
        cc_arr = cluster_centers_0(X, nClusters, random_state)
        cc_arr = np.ascontiguousarray(cc_arr, dtype=X.dtype)
        _validate_center_shape(X, nClusters, cc_arr)
        centroids_ = cc_arr
    elif is_string(cluster_centers_0, 'deterministic'):
        deterministic = True
        default_method = "lloydCSR" if is_sparse else "defaultDense"
        kmeans_init = daal4py.kmeans_init(
            nClusters, fptype=X_fptype, method=default_method)
        kmeans_init_res = kmeans_init.compute(X)
        centroids_ = kmeans_init_res.centroids
    else:
        raise ValueError(
            f"init should be either 'k-means++', 'random', a ndarray or a "
            f"callable, got '{cluster_centers_0}' instead.")
    if verbose:
        print("Initialization complete")
    return deterministic, centroids_


def _daal4py_kmeans_compatibility(nClusters, maxIterations, fptype="double",
                                  method="lloydDense", accuracyThreshold=0.0,
                                  resultsToEvaluate="computeCentroids"):
    kmeans_algo = daal4py.kmeans(
        nClusters=nClusters,
        maxIterations=maxIterations,
        fptype=fptype,
        resultsToEvaluate=resultsToEvaluate,
        accuracyThreshold=accuracyThreshold,
        method=method,
    )
    return kmeans_algo


def _daal4py_k_means_predict(X, nClusters, centroids,
                             resultsToEvaluate='computeAssignments'):
    X_fptype = getFPType(X)
    is_sparse = sp.isspmatrix(X)
    method = "lloydCSR" if is_sparse else "defaultDense"
    kmeans_algo = _daal4py_kmeans_compatibility(
        nClusters=nClusters,
        maxIterations=0,
        fptype=X_fptype,
        resultsToEvaluate=resultsToEvaluate,
        method=method,
    )

    res = kmeans_algo.compute(X, centroids)

    return res.assignments[:, 0], res.objectiveFunction[0, 0]


def _daal4py_k_means_fit(X, nClusters, numIterations,
                         tol, cluster_centers_0, n_init, verbose, random_state):
    if numIterations < 0:
        raise ValueError("Wrong iterations number")

    X_fptype = getFPType(X)
    abs_tol = _tolerance(X, tol)  # tol is relative tolerance
    is_sparse = sp.isspmatrix(X)
    method = "lloydCSR" if is_sparse else "defaultDense"
    best_inertia, best_cluster_centers = None, None
    best_n_iter = -1
    kmeans_algo = _daal4py_kmeans_compatibility(
        nClusters=nClusters,
        maxIterations=numIterations,
        accuracyThreshold=abs_tol,
        fptype=X_fptype,
        resultsToEvaluate='computeCentroids',
        method=method,
    )

    for k in range(n_init):
        deterministic, starting_centroids_ = _daal4py_compute_starting_centroids(
            X, X_fptype, nClusters, cluster_centers_0, verbose, random_state)

        res = kmeans_algo.compute(X, starting_centroids_)

        inertia = res.objectiveFunction[0, 0]
        if verbose:
            print(f"Iteration {k}, inertia {inertia}.")

        if best_inertia is None or inertia < best_inertia:
            best_cluster_centers = res.centroids
            if n_init > 1:
                best_cluster_centers = best_cluster_centers.copy()
            best_inertia = inertia
            best_n_iter = int(res.nIterations[0, 0])
        if deterministic and n_init != 1:
            warnings.warn(
                'Explicit initial center position passed: '
                'performing only one init in k-means instead of n_init=%d'
                % n_init, RuntimeWarning, stacklevel=2)
            break

    flag_compute = 'computeAssignments|computeExactObjectiveFunction'
    best_labels, best_inertia = _daal4py_k_means_predict(
        X, nClusters, best_cluster_centers, flag_compute)

    distinct_clusters = np.unique(best_labels).size
    if distinct_clusters < nClusters:
        warnings.warn(
            "Number of distinct clusters ({}) found smaller than "
            "n_clusters ({}). Possibly due to duplicate points "
            "in X.".format(distinct_clusters, nClusters),
            ConvergenceWarning, stacklevel=2)
        # for passing test case "test_kmeans_warns_less_centers_than_unique_points"

    return best_cluster_centers, best_labels, best_inertia, best_n_iter


def _fit(self, X, y=None, sample_weight=None):
    """Compute k-means clustering.

    Parameters
    ----------
    X : array-like or sparse matrix, shape=(n_samples, n_features)
        Training instances to cluster. It must be noted that the data
        will be converted to C ordering, which will cause a memory
        copy if the given data is not C-contiguous.

    y : Ignored
        not used, present here for API consistency by convention.

    sample_weight : array-like, shape (n_samples,), optional
        The weights for each observation in X. If None, all observations
        are assigned equal weight (default: None)

    """
    if hasattr(self, 'precompute_distances'):
        if self.precompute_distances != 'deprecated':
            if sklearn_check_version('0.24'):
                warnings.warn("'precompute_distances' was deprecated in version "
                              "0.23 and will be removed in 1.0 (renaming of 0.25)."
                              " It has no effect", FutureWarning)
            elif sklearn_check_version('0.23'):
                warnings.warn("'precompute_distances' was deprecated in version "
                              "0.23 and will be removed in 0.25. It has no "
                              "effect", FutureWarning)

    self._n_threads = None
    if hasattr(self, 'n_jobs'):
        if self.n_jobs != 'deprecated':
            if sklearn_check_version('0.24'):
                warnings.warn("'n_jobs' was deprecated in version 0.23 and will be"
                              " removed in 1.0 (renaming of 0.25).", FutureWarning)
            elif sklearn_check_version('0.23'):
                warnings.warn("'n_jobs' was deprecated in version 0.23 and will be"
                              " removed in 0.25.", FutureWarning)
            self._n_threads = self.n_jobs
    self._n_threads = _openmp_effective_n_threads(self._n_threads)

    if self.n_init <= 0:
        raise ValueError(
            f"n_init should be > 0, got {self.n_init} instead.")

    random_state = check_random_state(self.random_state)
    if sklearn_check_version("1.0"):
        self._check_feature_names(X, reset=True)

    if self.max_iter <= 0:
        raise ValueError(
            f"max_iter should be > 0, got {self.max_iter} instead.")

    algorithm = self.algorithm
    if algorithm == "elkan" and self.n_clusters == 1:
        warnings.warn("algorithm='elkan' doesn't make sense for a single "
                      "cluster. Using 'full' instead.", RuntimeWarning)
        algorithm = "full"

    if algorithm == "auto":
        algorithm = "full" if self.n_clusters == 1 else "elkan"

    if algorithm not in ["full", "elkan"]:
        raise ValueError("Algorithm must be 'auto', 'full' or 'elkan', got"
                         " {}".format(str(algorithm)))

    X_len = _num_samples(X)

    _patching_status = PatchingConditionsChain(
        "sklearn.cluster.KMeans.fit")
    _dal_ready = _patching_status.and_conditions([
        (self.n_clusters <= X_len,
            "The number of clusters is larger than the number of samples in X.")
    ])

    if _dal_ready and sample_weight is not None:
        if isinstance(sample_weight, numbers.Number):
            sample_weight = np.full(X_len, sample_weight, dtype=np.float64)
        else:
            sample_weight = np.asarray(sample_weight)
        _dal_ready = _patching_status.and_conditions([
            (sample_weight.shape == (X_len,),
                "Sample weights do not have the same length as X."),
            (np.allclose(sample_weight, np.ones_like(sample_weight)),
                "Sample weights are not ones.")
        ])

    _patching_status.write_log()
    if _dal_ready:
        X = check_array(X, accept_sparse='csr', dtype=[np.float64, np.float32])
        self.n_features_in_ = X.shape[1]
        self.cluster_centers_, self.labels_, self.inertia_, self.n_iter_ = \
            _daal4py_k_means_fit(
                X, self.n_clusters, self.max_iter, self.tol, self.init, self.n_init,
                self.verbose, random_state)
    else:
        super(KMeans, self).fit(X, y=y, sample_weight=sample_weight)
    return self


def _daal4py_check_test_data(self, X):
    if sklearn_check_version("1.0"):
        self._check_feature_names(X, reset=False)
    X = check_array(
        X,
        accept_sparse='csr',
        dtype=[np.float64, np.float32],
        accept_large_sparse=False
    )
    if self.n_features_in_ != X.shape[1]:
        raise ValueError(
            (f'X has {X.shape[1]} features, '
             f'but Kmeans is expecting {self.n_features_in_} features as input'))
    return X


def _predict(self, X, sample_weight=None):
    """Predict the closest cluster each sample in X belongs to.

    In the vector quantization literature, `cluster_centers_` is called
    the code book and each value returned by `predict` is the index of
    the closest code in the code book.

    Parameters
    ----------
    X : {array-like, sparse matrix}, shape = [n_samples, n_features]
       New data to predict.

    sample_weight : array-like, shape (n_samples,), optional
        The weights for each observation in X. If None, all observations
        are assigned equal weight (default: None)

    Returns
    -------
    labels : array, shape [n_samples,]
        Index of the cluster each sample belongs to.
    """
    check_is_fitted(self)

    X = _daal4py_check_test_data(self, X)

    _patching_status = PatchingConditionsChain(
        "sklearn.cluster.KMeans.predict")
    _patching_status.and_conditions([
        (sample_weight is None, "Sample weights are not supported."),
        (hasattr(X, '__array__'), "X does not have '__array__' attribute.")
    ])
    _dal_ready = _patching_status.or_conditions([
        (sp.isspmatrix_csr(X), "X is not sparse.")
    ])

    _patching_status.write_log()
    if _dal_ready:
        return _daal4py_k_means_predict(
            X, self.n_clusters, self.cluster_centers_)[0]
    x_squared_norms = row_norms(X, squared=True)
    return _labels_inertia(X, sample_weight, x_squared_norms,
                           self.cluster_centers_)[0]


class KMeans(KMeans_original):
    __doc__ = KMeans_original.__doc__

    if sklearn_check_version('1.0'):
        @_deprecate_positional_args
        def __init__(
            self,
            n_clusters=8,
            *,
            init='k-means++',
            n_init=10,
            max_iter=300,
            tol=1e-4,
            verbose=0,
            random_state=None,
            copy_x=True,
            algorithm='auto',
        ):
            super(KMeans, self).__init__(
                n_clusters=n_clusters,
                init=init,
                max_iter=max_iter,
                tol=tol,
                n_init=n_init,
                verbose=verbose,
                random_state=random_state,
                copy_x=copy_x,
                algorithm=algorithm,
            )
    else:
        @_deprecate_positional_args
        def __init__(
            self,
            n_clusters=8,
            *,
            init='k-means++',
            n_init=10,
            max_iter=300,
            tol=1e-4,
            precompute_distances='deprecated',
            verbose=0,
            random_state=None,
            copy_x=True,
            n_jobs='deprecated',
            algorithm='auto',
        ):
            super(KMeans, self).__init__(
                n_clusters=n_clusters,
                init=init,
                max_iter=max_iter,
                tol=tol,
                precompute_distances=precompute_distances,
                n_init=n_init,
                verbose=verbose,
                random_state=random_state,
                copy_x=copy_x,
                n_jobs=n_jobs,
                algorithm=algorithm,
            )

    @support_usm_ndarray()
    def fit(self, X, y=None, sample_weight=None):
        """
        Compute k-means clustering.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Training instances to cluster. It must be noted that the data
            will be converted to C ordering, which will cause a memory
            copy if the given data is not C-contiguous.
            If a sparse matrix is passed, a copy will be made if it's not in
            CSR format.

        y : Ignored
            Not used, present here for API consistency by convention.

        sample_weight : array-like of shape (n_samples,), default=None
            The weights for each observation in X. If None, all observations
            are assigned equal weight.

            .. versionadded:: 0.20

        Returns
        -------
        self : object
            Fitted estimator.
        """
        return _fit(self, X, y=y, sample_weight=sample_weight)

    @support_usm_ndarray()
    def predict(self, X, sample_weight=None):
        """
        Predict the closest cluster each sample in X belongs to.

        In the vector quantization literature, `cluster_centers_` is called
        the code book and each value returned by `predict` is the index of
        the closest code in the code book.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            New data to predict.

        sample_weight : array-like of shape (n_samples,), default=None
            The weights for each observation in X. If None, all observations
            are assigned equal weight.

        Returns
        -------
        labels : ndarray of shape (n_samples,)
            Index of the cluster each sample belongs to.
        """
        return _predict(self, X, sample_weight=sample_weight)

    @support_usm_ndarray()
    def fit_predict(self, X, y=None, sample_weight=None):
        """
        Compute cluster centers and predict cluster index for each sample.

        Convenience method; equivalent to calling fit(X) followed by
        predict(X).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            New data to transform.

        y : Ignored
            Not used, present here for API consistency by convention.

        sample_weight : array-like of shape (n_samples,), default=None
            The weights for each observation in X. If None, all observations
            are assigned equal weight.

        Returns
        -------
        labels : ndarray of shape (n_samples,)
            Index of the cluster each sample belongs to.
        """
        return super().fit_predict(X, y, sample_weight)



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# daal4py logistic regression example for shared memory systems

import daal4py as d4p
import numpy as np

# let's try to use pandas' fast csv reader
try:
    import pandas

    def read_csv(f, c, t=np.float64):
        return pandas.read_csv(f, usecols=c, delimiter=',', header=None, dtype=t)
except ImportError:
    # fall back to numpy loadtxt
    def read_csv(f, c, t=np.float64):
        return np.loadtxt(f, usecols=c, delimiter=',', ndmin=2)


def main(readcsv=read_csv, method='defaultDense'):
    nClasses = 2
    nFeatures = 20

    # read training data from file with 20 features per observation and 1 class label
    trainfile = "./data/batch/binary_cls_train.csv"
    train_data = readcsv(trainfile, range(nFeatures))
    train_labels = readcsv(trainfile, range(nFeatures, nFeatures + 1))

    # set parameters and train
    train_alg = d4p.logistic_regression_training(nClasses=nClasses, interceptFlag=True)
    train_result = train_alg.compute(train_data, train_labels)

    # read testing data from file with 20 features per observation
    testfile = "./data/batch/binary_cls_test.csv"
    predict_data = readcsv(testfile, range(nFeatures))
    predict_labels = readcsv(testfile, range(nFeatures, nFeatures + 1))

    # set parameters and compute predictions
    predict_alg = d4p.logistic_regression_prediction(nClasses=nClasses)
    predict_result = predict_alg.compute(predict_data, train_result.model)

    # the prediction result provides prediction
    assert predict_result.prediction.shape == (predict_data.shape[0],
                                               train_labels.shape[1])

    return (train_result, predict_result, predict_labels)


if __name__ == "__main__":
    (train_result, predict_result, predict_labels) = main()
    print("\nLogistic Regression coefficients:\n", train_result.model.Beta)
    print(
        "\nLogistic regression prediction results (first 10 rows):\n",
        predict_result.prediction[0:10]
    )
    print("\nGround truth (first 10 rows):\n", predict_labels[0:10])
    print('All looks good!')



#===============================================================================
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import pytest
import functools


def get_queues(filter_='cpu,gpu,host'):
    queues = []

    if 'host' in filter_:
        queues.append(None)

    try:
        import dpctl

        if dpctl.has_cpu_devices and 'cpu' in filter_:
            queues.append(dpctl.SyclQueue('cpu'))
        if dpctl.has_gpu_devices and 'gpu' in filter_:
            queues.append(dpctl.SyclQueue('gpu'))
    finally:
        return queues


def get_memory_usm():
    try:
        from dpctl.memory import MemoryUSMDevice, MemoryUSMShared
        return [MemoryUSMDevice, MemoryUSMShared]
    except ImportError:
        return []


def is_dpctl_available(targets=None):
    try:
        import dpctl

        if targets is None:
            return True
        for device in targets:
            if device == 'cpu' and not dpctl.has_cpu_devices():
                return False
            if device == 'gpu' and not dpctl.has_gpu_devices():
                return False
        return True
    except ImportError:
        return False


def device_type_to_str(queue):
    if queue is None:
        return 'host'

    from dpctl import device_type
    if queue.sycl_device.device_type == device_type.cpu:
        return 'cpu'
    if queue.sycl_device.device_type == device_type.gpu:
        return 'gpu'
    if queue.sycl_device.device_type == device_type.host:
        return 'host'
    return 'unknown'


def pass_if_not_implemented_for_gpu(reason=""):
    assert reason

    def decorator(test):
        @functools.wraps(test)
        def wrapper(queue, *args, **kwargs):
            if queue is not None and queue.sycl_device.is_gpu:
                with pytest.raises(RuntimeError, match='is not implemented for GPU'):
                    test(queue, *args, **kwargs)
            else:
                test(queue, *args, **kwargs)
        return wrapper
    return decorator



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np
import sys

from daal4py import _get__daal_link_version__ as dv
from sklearn import __version__ as sklearn_version
try:
    from packaging.version import Version
except ImportError:
    from distutils.version import LooseVersion as Version
import logging


def set_idp_sklearn_verbose():
    import warnings
    import os
    logLevel = os.environ.get("IDP_SKLEARN_VERBOSE")
    try:
        if logLevel is not None:
            logging.basicConfig(
                stream=sys.stdout,
                format='%(levelname)s: %(message)s', level=logLevel.upper())
    except Exception:
        warnings.warn('Unknown level "{}" for logging.\n'
                      'Please, use one of "CRITICAL", "ERROR", '
                      '"WARNING", "INFO", "DEBUG".'.format(logLevel))


def daal_check_version(rule):
    # First item is major version - 2021,
    # second is minor+patch - 0110,
    # third item is status - B
    target = (int(dv()[0:4]), dv()[10:11], int(dv()[4:8]))
    if not isinstance(rule[0], type(target)):
        if rule > target:
            return False
    else:
        for rule_item in rule:
            if rule_item > target:
                return False
            if rule_item[0] == target[0]:
                break
    return True


def sklearn_check_version(ver):
    if hasattr(Version(ver), 'base_version'):
        base_sklearn_version = Version(sklearn_version).base_version
        return bool(Version(base_sklearn_version) >= Version(ver))
    # packaging module not available
    return bool(Version(sklearn_version) >= Version(ver))


def get_daal_version():
    return (int(dv()[0:4]), dv()[10:11], int(dv()[4:8]))


def parse_dtype(dt):
    if dt == np.double:
        return "double"
    if dt == np.single:
        return "float"
    raise ValueError(f"Input array has unexpected dtype = {dt}")


def getFPType(X):
    try:
        from pandas import DataFrame
        from pandas.core.dtypes.cast import find_common_type
        if isinstance(X, DataFrame):
            dt = find_common_type(X.dtypes.tolist())
            return parse_dtype(dt)
    except ImportError:
        pass

    dt = getattr(X, 'dtype', None)
    return parse_dtype(dt)


def make2d(X):
    if np.isscalar(X):
        X = np.asarray(X)[np.newaxis, np.newaxis]
    elif isinstance(X, np.ndarray) and X.ndim == 1:
        X = X.reshape((X.size, 1))
    return X


def get_patch_message(s):
    if s == "daal":
        message = "running accelerated version on "
        if 'daal4py.oneapi' in sys.modules:
            from daal4py.oneapi import _get_device_name_sycl_ctxt
            dev = _get_device_name_sycl_ctxt()
            if dev == 'cpu' or dev == 'host' or dev is None:
                message += 'CPU'
            elif dev == 'gpu':
                message += 'GPU'
            else:
                raise ValueError(f"Unexpected device name {dev}."
                                 " Supported types are host, cpu and gpu")
        else:
            message += 'CPU'

    elif s == "sklearn":
        message = "fallback to original Scikit-learn"
    elif s == "sklearn_after_daal":
        message = "failed to run accelerated version, fallback to original Scikit-learn"
    else:
        raise ValueError(
            f"Invalid input - expected one of 'daal','sklearn',"
            f" 'sklearn_after_daal', got {s}")
    return message


def is_in_sycl_ctxt():
    try:
        from daal4py.oneapi import is_in_sycl_ctxt as is_in_ctx
        return is_in_ctx()
    except ModuleNotFoundError:
        return False


def is_DataFrame(X):
    try:
        from pandas import DataFrame
        return isinstance(X, DataFrame)
    except ImportError:
        return False


def get_dtype(X):
    try:
        from pandas.core.dtypes.cast import find_common_type
        return find_common_type(list(X.dtypes)) if is_DataFrame(X) else X.dtype
    except ImportError:
        return getattr(X, "dtype", None)


def get_number_of_types(dataframe):
    dtypes = getattr(dataframe, "dtypes", None)
    try:
        return len(set(dtypes))
    except TypeError:
        return 1


class PatchingConditionsChain:
    def __init__(self, scope_name):
        self.scope_name = scope_name
        self.patching_is_enabled = True
        self.messages = []

    def _iter_conditions(self, conditions_and_messages):
        result = []
        for condition, message in conditions_and_messages:
            result.append(condition)
            if not condition:
                self.messages.append(message)
        return result

    def and_conditions(self, conditions_and_messages, conditions_merging=all):
        self.patching_is_enabled &= conditions_merging(
            self._iter_conditions(conditions_and_messages))
        return self.patching_is_enabled

    def or_conditions(self, conditions_and_messages, conditions_merging=all):
        self.patching_is_enabled |= conditions_merging(
            self._iter_conditions(conditions_and_messages))
        return self.patching_is_enabled

    def get_status(self):
        return self.patching_is_enabled

    def write_log(self):
        if self.patching_is_enabled:
            logging.info(f"{self.scope_name}: {get_patch_message('daal')}")
        else:
            logging.debug(
                f'{self.scope_name}: debugging for the patch is enabled to track'
                ' the usage of Intel® oneAPI Data Analytics Library (oneDAL)')
            for message in self.messages:
                logging.debug(
                    f'{self.scope_name}: patching failed with cause - {message}')
            logging.info(f"{self.scope_name}: {get_patch_message('sklearn')}")



#===============================================================================
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# daal4py Gradient Bossting Classification model creation from Catboost example

import daal4py as d4p
import catboost as cb
import numpy as np
import pandas as pd


def pd_read_csv(f, c=None, t=np.float64):
    return pd.read_csv(f, usecols=c, delimiter=',', header=None, dtype=t)


def main(readcsv=pd_read_csv, method='defaultDense'):
    # Path to data
    train_file = "./data/batch/df_classification_train.csv"
    test_file = "./data/batch/df_classification_test.csv"

    # Data reading
    X_train = readcsv(train_file, range(3), t=np.float32)
    y_train = readcsv(train_file, range(3, 4), t=np.float32)
    X_test = readcsv(test_file, range(3), t=np.float32)
    y_test = readcsv(test_file, range(3, 4), t=np.float32)

    # Datasets creation
    cb_train = cb.Pool(X_train, label=np.array(y_train))
    cb_test = cb.Pool(X_test, label=np.array(y_test))

    # training parameters setting
    params = {
        'reg_lambda': 1,
        'max_depth': 8,
        'num_leaves': 2**8,
        'verbose': 0,
        'objective': 'MultiClass',
        'learning_rate': 0.3,
        'n_estimators': 100,
        'classes_count': 5,
    }

    # Training
    cb_model = cb.CatBoost(params)
    cb_model.fit(cb_train)

    # Catboost prediction
    cb_prediction = cb_model.predict(cb_test, prediction_type='Class').T[0]
    cb_errors_count = np.count_nonzero(cb_prediction - np.ravel(y_test))

    # Conversion to daal4py
    daal_model = d4p.get_gbt_model_from_catboost(cb_model)

    # daal4py prediction
    daal_predict_algo = d4p.gbt_classification_prediction(
        nClasses=params['classes_count'],
        resultsToEvaluate="computeClassLabels",
        fptype='float'
    )
    daal_prediction = daal_predict_algo.compute(X_test, daal_model)
    daal_errors_count = np.count_nonzero(daal_prediction.prediction - y_test)
    assert np.absolute(cb_errors_count - daal_errors_count) == 0

    return (cb_prediction, cb_errors_count, np.ravel(daal_prediction.prediction),
            daal_errors_count, np.ravel(y_test))


if __name__ == "__main__":
    (cb_prediction, cb_errors_count,
     daal_prediction, daal_errors_count, y_test) = main()
    print("\nCatboost prediction results (first 10 rows):\n",
          cb_prediction[0:10])
    print("\ndaal4py prediction results (first 10 rows):\n",
          daal_prediction[0:10])
    print("\nGround truth (first 10 rows):\n", y_test[0:10])

    print("Catboost errors count:", cb_errors_count)
    print("Catboost accuracy score:",
          1 - cb_errors_count / cb_prediction.shape[0])

    print("\ndaal4py errors count:", daal_errors_count)
    print("daal4py accuracy score:",
          1 - daal_errors_count / daal_prediction.shape[0])
    print("\nAll looks good!")



#===============================================================================
# Copyright 2020 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np
import pytest
import random
from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split
from sklearn.metrics import (accuracy_score, roc_auc_score,
                             mean_squared_error, log_loss)
from sklearn.ensemble \
    import RandomForestClassifier as ScikitRandomForestClassifier
from daal4py.sklearn.ensemble \
    import RandomForestClassifier as DaalRandomForestClassifier
from sklearn.ensemble \
    import RandomForestRegressor as ScikitRandomForestRegressor
from daal4py.sklearn.ensemble \
    import RandomForestRegressor as DaalRandomForestRegressor
from daal4py.sklearn._utils import daal_check_version

ACCURACY_RATIO = 0.95 if daal_check_version((2021, 'P', 400)) else 0.85
MSE_RATIO = 1.07
LOG_LOSS_RATIO = 1.4 if daal_check_version((2021, 'P', 400)) else 1.55
ROC_AUC_RATIO = 0.96
RNG = np.random.RandomState(0)
IRIS = load_iris()


def _compare_with_sklearn_classifier_iris(n_estimators=100, class_weight=None,
                                          sample_weight=None, description=""):
    x_train, x_test, y_train, y_test = \
        train_test_split(IRIS.data, IRIS.target,
                         test_size=0.33, random_state=31)
    # models
    scikit_model = ScikitRandomForestClassifier(n_estimators=n_estimators,
                                                class_weight=class_weight,
                                                max_depth=None,
                                                random_state=777)
    daal4py_model = DaalRandomForestClassifier(n_estimators=n_estimators,
                                               class_weight=class_weight,
                                               max_depth=None,
                                               random_state=777)
    # training
    scikit_model.fit(x_train, y_train, sample_weight=sample_weight)
    daal4py_model.fit(x_train, y_train, sample_weight=sample_weight)
    #predict
    scikit_predict = scikit_model.predict(x_test)
    daal4py_predict = daal4py_model.predict(x_test)
    #accuracy
    scikit_accuracy = accuracy_score(scikit_predict, y_test)
    daal4py_accuracy = accuracy_score(daal4py_predict, y_test)
    ratio = daal4py_accuracy / scikit_accuracy
    reason = description + \
        f"scikit_accuracy={scikit_accuracy}, daal4py_accuracy={daal4py_accuracy}"
    assert ratio >= ACCURACY_RATIO, reason

    # predict_proba
    scikit_predict_proba = scikit_model.predict_proba(x_test)
    daal4py_predict_proba = daal4py_model.predict_proba(x_test)
    #log_loss
    scikit_log_loss = log_loss(y_test, scikit_predict_proba)
    daal4py_log_loss = log_loss(y_test, daal4py_predict_proba)
    ratio = daal4py_log_loss / scikit_log_loss

    reason = description +\
        f"scikit_log_loss={scikit_log_loss}, daal4py_log_loss={daal4py_log_loss}"
    assert ratio <= LOG_LOSS_RATIO, reason

    # ROC AUC
    scikit_roc_auc = roc_auc_score(y_test, scikit_predict_proba, multi_class='ovr')
    daal4py_roc_auc = roc_auc_score(y_test, daal4py_predict_proba, multi_class='ovr')
    ratio = daal4py_roc_auc / scikit_roc_auc

    reason = description + \
        f"scikit_roc_auc={scikit_roc_auc}, daal4py_roc_auc={daal4py_roc_auc}"
    assert ratio >= ROC_AUC_RATIO, reason


CLASS_WEIGHTS_IRIS = [
    {0: 0, 1: 1, 2: 1},
    {0: 1, 1: 2, 2: 3},
    {0: 10, 1: 5, 2: 4},
    {
        0: RNG.uniform(1, 50),
        1: RNG.uniform(1, 50),
        2: RNG.uniform(1, 50),
    },
    {
        0: RNG.uniform(1, 1000),
        1: RNG.uniform(1, 1000),
        2: RNG.uniform(1, 1000),
    },
    {
        0: RNG.uniform(1, 10),
        1: RNG.uniform(50, 100),
        2: RNG.uniform(1, 100),
    },
    {0: 50, 1: 50, 2: 50},
    'balanced',
]


@pytest.mark.parametrize('class_weight', CLASS_WEIGHTS_IRIS)
def test_classifier_class_weight_iris(class_weight):
    _compare_with_sklearn_classifier_iris(
        class_weight=class_weight,
        description='Classifier class weight: '
    )


SAMPLE_WEIGHTS_IRIS = [
    (np.full_like(range(100), 1), 'Only 1'),
    (np.full_like(range(100), 50), 'Only 50'),
    (RNG.rand(100), 'Uniform distribution'),
    (RNG.normal(1000, 10, 100), 'Gaussian distribution'),
    (RNG.poisson(lam=10, size=100), 'Poisson distribution'),
    (RNG.rayleigh(scale=1, size=100), 'Rayleigh distribution'),
]


@pytest.mark.parametrize('sample_weight', SAMPLE_WEIGHTS_IRIS)
def test_classifier_sample_weight_iris(sample_weight):
    sample_weight, description = sample_weight
    _compare_with_sklearn_classifier_iris(
        sample_weight=sample_weight,
        description=f'Classifier sample_weight_type={description}: '
    )


N_ESTIMATORS_IRIS = [
    1000,
    8000,
]


@pytest.mark.parametrize('n_estimators', N_ESTIMATORS_IRIS)
def test_classifier_big_estimators_iris(n_estimators):
    _compare_with_sklearn_classifier_iris(
        n_estimators=n_estimators,
        description=f'Classifier n_estimators={n_estimators}: '
    )


def _compare_with_sklearn_mse_regressor_iris(n_estimators=100, sample_weight=None,
                                             description=""):
    x_train, x_test, y_train, y_test = \
        train_test_split(IRIS.data, IRIS.target,
                         test_size=0.33, random_state=31)

    scikit_model = ScikitRandomForestRegressor(n_estimators=n_estimators,
                                               max_depth=None,
                                               random_state=777)
    daal4py_model = DaalRandomForestRegressor(n_estimators=n_estimators,
                                              max_depth=None,
                                              random_state=777)

    scikit_predict = scikit_model.fit(
        x_train, y_train,
        sample_weight=sample_weight).predict(x_test)
    daal4py_predict = daal4py_model.fit(
        x_train, y_train,
        sample_weight=sample_weight).predict(x_test)

    scikit_mse = mean_squared_error(scikit_predict, y_test)
    daal4py_mse = mean_squared_error(daal4py_predict, y_test)

    ratio = daal4py_mse / scikit_mse
    reason = description + f"scikit_mse={scikit_mse}, daal4py_mse={daal4py_mse}"
    assert ratio <= MSE_RATIO, reason


@pytest.mark.parametrize('weight', SAMPLE_WEIGHTS_IRIS)
def test_mse_regressor_sample_weight_iris(weight):
    sample_weight, description = weight
    _compare_with_sklearn_mse_regressor_iris(
        sample_weight=sample_weight,
        description=f"Regression sample weights: sample_weight_type={description}: "
    )


@pytest.mark.parametrize('n_estimators', N_ESTIMATORS_IRIS)
def test_mse_regressor_big_estimators_iris(n_estimators):
    _compare_with_sklearn_mse_regressor_iris(
        n_estimators=n_estimators,
        description=f"Regression: n_estimators={n_estimators}: "
    )



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np
from scipy import sparse as sp
from sklearn.utils import check_array, check_X_y
from sklearn.linear_model._ridge import _BaseRidge
from sklearn.linear_model._ridge import Ridge as Ridge_original

import daal4py
from .._utils import (
    make2d, getFPType, get_patch_message, sklearn_check_version,
    PatchingConditionsChain)
from .._device_offload import support_usm_ndarray
import logging


def _daal4py_fit(self, X, y_):
    X = make2d(X)
    y = make2d(y_)

    _fptype = getFPType(X)

    ridge_params = np.asarray(self.alpha, dtype=X.dtype)
    if ridge_params.size != 1 and ridge_params.size != y.shape[1]:
        raise ValueError("alpha length is wrong")
    ridge_params = ridge_params.reshape((1, -1))

    ridge_alg = daal4py.ridge_regression_training(
        fptype=_fptype,
        method='defaultDense',
        interceptFlag=(self.fit_intercept is True),
        ridgeParameters=ridge_params
    )
    try:
        ridge_res = ridge_alg.compute(X, y)
    except RuntimeError:
        return None

    ridge_model = ridge_res.model
    self.daal_model_ = ridge_model
    coefs = ridge_model.Beta

    self.intercept_ = coefs[:, 0].copy(order='C')
    self.coef_ = coefs[:, 1:].copy(order='C')

    if self.coef_.shape[0] == 1 and y_.ndim == 1:
        self.coef_ = np.ravel(self.coef_)
        self.intercept_ = self.intercept_[0]

    return self


def _daal4py_predict(self, X):
    X = make2d(X)
    _fptype = getFPType(self.coef_)

    ridge_palg = daal4py.ridge_regression_prediction(
        fptype=_fptype,
        method='defaultDense'
    )
    if self.n_features_in_ != X.shape[1]:
        raise ValueError(
            f'X has {X.shape[1]} features, '
            f'but Ridge is expecting {self.n_features_in_} features as input'
        )
    ridge_res = ridge_palg.compute(X, self.daal_model_)

    res = ridge_res.prediction

    if res.shape[1] == 1 and self.coef_.ndim == 1:
        res = np.ravel(res)
    return res


def _fit_ridge(self, X, y, sample_weight=None):
    """Fit Ridge regression model

    Parameters
    ----------
    X : {array-like, sparse matrix}, shape = [n_samples, n_features]
        Training data

    y : array-like, shape = [n_samples] or [n_samples, n_targets]
        Target values

    sample_weight : float or numpy array of shape [n_samples]
        Individual weights for each sample

    Returns
    -------
    self : returns an instance of self.
    """
    if sklearn_check_version('1.0'):
        from sklearn.linear_model._base import _deprecate_normalize
        self._normalize = _deprecate_normalize(
            self.normalize,
            default=False,
            estimator_name=self.__class__.__name__
        )
        self._check_feature_names(X, reset=True)

    X, y = check_X_y(X, y, ['csr', 'csc', 'coo'], dtype=[np.float64, np.float32],
                     multi_output=True, y_numeric=True)
    self.n_features_in_ = X.shape[1]
    self.sample_weight_ = sample_weight
    self.fit_shape_good_for_daal_ = True if X.shape[0] >= X.shape[1] else False

    _patching_status = PatchingConditionsChain(
        "sklearn.linear_model.Ridge.fit")
    _dal_ready = _patching_status.and_conditions([
        (self.solver == 'auto',
            f"'{self.solver}' solver is not supported. "
            "Only 'auto' solver is supported."),
        (not sp.issparse(X), "X is sparse. Sparse input is not supported."),
        (self.fit_shape_good_for_daal_,
            "The shape of X does not satisfy oneDAL requirements: "
            "number of features > number of samples."),
        (X.dtype == np.float64 or X.dtype == np.float32,
            f"'{X.dtype}' X data type is not supported. "
            "Only np.float32 and np.float64 are supported."),
        (sample_weight is None, "Sample weights are not supported."),
        (not (hasattr(self, 'positive') and self.positive),
            "Forced positive coefficients are not supported.")])
    _patching_status.write_log()

    if not _dal_ready:
        if hasattr(self, 'daal_model_'):
            del self.daal_model_
        return super(Ridge, self).fit(X, y, sample_weight=sample_weight)
    self.n_iter_ = None
    res = _daal4py_fit(self, X, y)
    if res is None:
        logging.info(
            "sklearn.linear_model.Ridge.fit: " + get_patch_message("sklearn_after_daal"))
        if hasattr(self, 'daal_model_'):
            del self.daal_model_
        return super(Ridge, self).fit(X, y, sample_weight=sample_weight)
    return res


def _predict_ridge(self, X):
    """Predict using the linear model

    Parameters
    ----------
    X : {array-like, sparse matrix}, shape = (n_samples, n_features)
        Samples.

    Returns
    -------
    C : array, shape = (n_samples,)
        Returns predicted values.
    """
    if sklearn_check_version('1.0'):
        self._check_feature_names(X, reset=False)

    X = check_array(
        X, accept_sparse=['csr', 'csc', 'coo'], dtype=[np.float64, np.float32])
    good_shape_for_daal = \
        True if X.ndim <= 1 else True if X.shape[0] >= X.shape[1] else False

    _patching_status = PatchingConditionsChain(
        "sklearn.linear_model.Ridge.predict")
    _dal_ready = _patching_status.and_conditions([
        (self.solver == 'auto',
            f"'{self.solver}' solver is not supported. "
            "Only 'auto' solver is supported."),
        (hasattr(self, 'daal_model_'), "oneDAL model was not trained."),
        (not sp.issparse(X), "X is sparse. Sparse input is not supported."),
        (good_shape_for_daal,
            "The shape of X does not satisfy oneDAL requirements: "
            "number of features > number of samples."),
        (X.dtype == np.float64 or X.dtype == np.float32,
            f"'{X.dtype}' X data type is not supported. "
            "Only np.float32 and np.float64 are supported."),
        (not hasattr(self, 'sample_weight_') or self.sample_weight_ is None,
            "Sample weights are not supported.")])
    _patching_status.write_log()

    if not _dal_ready:
        return self._decision_function(X)
    return _daal4py_predict(self, X)


class Ridge(Ridge_original, _BaseRidge):
    __doc__ = Ridge_original.__doc__

    if sklearn_check_version('1.0'):
        def __init__(
            self,
            alpha=1.0,
            fit_intercept=True,
            normalize='deprecated',
            copy_X=True,
            max_iter=None,
            tol=1e-3,
            solver="auto",
            positive=False,
            random_state=None,
        ):
            self.alpha = alpha
            self.fit_intercept = fit_intercept
            self.normalize = normalize
            self.copy_X = copy_X
            self.max_iter = max_iter
            self.tol = tol
            self.solver = solver
            self.positive = positive
            self.random_state = random_state
    else:
        def __init__(
            self,
            alpha=1.0,
            fit_intercept=True,
            normalize=False,
            copy_X=True,
            max_iter=None,
            tol=1e-3,
            solver="auto",
            random_state=None,
        ):
            self.alpha = alpha
            self.fit_intercept = fit_intercept
            self.normalize = normalize
            self.copy_X = copy_X
            self.max_iter = max_iter
            self.tol = tol
            self.solver = solver
            self.random_state = random_state

    @support_usm_ndarray()
    def fit(self, X, y, sample_weight=None):
        """
        Fit Ridge regression model.

        Parameters
        ----------
        X : {ndarray, sparse matrix} of shape (n_samples, n_features)
            Training data.

        y : ndarray of shape (n_samples,) or (n_samples, n_targets)
            Target values.

        sample_weight : float or ndarray of shape (n_samples,), default=None
            Individual weights for each sample. If given a float, every sample
            will have the same weight.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        return _fit_ridge(self, X, y, sample_weight=sample_weight)

    @support_usm_ndarray()
    def predict(self, X):
        """
        Predict using the linear model.

        Parameters
        ----------
        X : array-like or sparse matrix, shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array, shape (n_samples,)
            Returns predicted values.
        """
        return _predict_ridge(self, X)



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# daal4py QR example for shared memory systems

import daal4py as d4p

# let's use a generator for getting stream from file (defined in stream.py)
from stream import read_next


def main(readcsv=None, method='svdDense'):
    infile = "./data/batch/qr.csv"

    # configure a QR object
    algo = d4p.qr(streaming=True)

    # get the generator (defined in stream.py)...
    rn = read_next(infile, 112, readcsv)
    # ... and iterate through chunks/stream
    for chunk in rn:
        algo.compute(chunk)

    # finalize computation
    result = algo.finalize()

    # QR result objects provide matrixQ and matrixR
    return result


if __name__ == "__main__":
    result = main()
    print("Orthogonal matrix Q:\n", result.matrixQ[:10])
    print("Triangular matrix R:\n", result.matrixR)
    print('All looks good!')



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# daal4py Adaboost example for shared memory systems

import daal4py as d4p
import numpy as np

# let's try to use pandas' fast csv reader
try:
    import pandas

    def read_csv(f, c, t=np.float64):
        return pandas.read_csv(f, usecols=c, delimiter=',', header=None, dtype=t)
except ImportError:
    # fall back to numpy loadtxt
    def read_csv(f, c, t=np.float64):
        return np.loadtxt(f, usecols=c, delimiter=',', ndmin=2)


def main(readcsv=read_csv, method='defaultDense'):
    infile = "./data/batch/adaboost_train.csv"
    testfile = "./data/batch/adaboost_test.csv"
    nClasses = 2

    # Configure a adaboost training object
    train_algo = d4p.adaboost_training(nClasses=nClasses)

    # Read data. Let's have 20 independent,
    # and 1 dependent variable (for each observation)
    indep_data = readcsv(infile, range(20))
    dep_data = readcsv(infile, range(20, 21))
    # Now train/compute, the result provides the model for prediction
    train_result = train_algo.compute(indep_data, dep_data)

    # Now let's do some prediction
    predict_algo = d4p.adaboost_prediction(nClasses=nClasses)
    # read test data (with same #features)
    pdata = readcsv(testfile, range(20))
    # now predict using the model from the training above
    predict_result = predict_algo.compute(pdata, train_result.model)

    # The prediction result provides prediction
    assert predict_result.prediction.shape == (pdata.shape[0], dep_data.shape[1])
    ptdata = np.loadtxt(testfile, usecols=range(20, 21), delimiter=',', ndmin=2)
    assert np.allclose(predict_result.prediction, ptdata)

    return (train_result, predict_result, ptdata)


if __name__ == "__main__":
    (train_result, predict_result, ptdata) = main()
    print("\nGround truth (first 20 observations):\n", ptdata[:20])
    print(
        "Adaboost classification results: (first 20 observations):\n",
        predict_result.prediction[:20]
    )
    print('All looks good!')



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np

import daal4py
from .._utils import (make2d, getFPType)


def _resultsToCompute_string(value=True, gradient=True, hessian=False):
    results_needed = []
    if value:
        results_needed.append('value')
    if gradient:
        results_needed.append('gradient')
    if hessian:
        results_needed.append('hessian')

    return '|'.join(results_needed)


def _daal4py_logistic_loss_extra_args(nClasses_unused, beta, X, y,
                                      l1=0.0, l2=0.0, fit_intercept=True,
                                      value=True, gradient=True, hessian=False):
    X = make2d(X)
    nSamples, nFeatures = X.shape

    y = make2d(y)
    beta = make2d(beta)
    n = X.shape[0]

    results_to_compute = _resultsToCompute_string(
        value=value, gradient=gradient, hessian=hessian)

    objective_function_algorithm_instance = \
        daal4py.optimization_solver_logistic_loss(
            numberOfTerms=n,
            fptype=getFPType(X),
            method='defaultDense',
            interceptFlag=fit_intercept,
            penaltyL1=l1 / n,
            penaltyL2=l2 / n,
            resultsToCompute=results_to_compute
        )
    objective_function_algorithm_instance.setup(X, y, beta)

    return (objective_function_algorithm_instance, X, y, n)


def _daal4py_cross_entropy_loss_extra_args(nClasses, beta, X, y,
                                           l1=0.0, l2=0.0, fit_intercept=True,
                                           value=True, gradient=True, hessian=False):
    X = make2d(X)
    nSamples, nFeatures = X.shape
    y = make2d(y)
    beta = make2d(beta)
    n = X.shape[0]

    results_to_compute = _resultsToCompute_string(
        value=value, gradient=gradient, hessian=hessian)

    objective_function_algorithm_instance = \
        daal4py.optimization_solver_cross_entropy_loss(
            nClasses=nClasses,
            numberOfTerms=n,
            fptype=getFPType(X),
            method='defaultDense',
            interceptFlag=fit_intercept,
            penaltyL1=l1 / n,
            penaltyL2=l2 / n,
            resultsToCompute=results_to_compute
        )
    objective_function_algorithm_instance.setup(X, y, beta)

    return (objective_function_algorithm_instance, X, y, n)


# used by LBFGS solver
def _daal4py_loss_and_grad(beta, objF_instance, X, y, n):
    beta_ = make2d(beta)
    res = objF_instance.compute(X, y, beta_)
    gr = res.gradientIdx.ravel()
    gr = gr * n  # force copy
    v = res.valueIdx[0, 0]
    v *= n
    return (v, gr)


# used by Newton CG method
def _daal4py_loss_(beta, objF_instance, X, y, n, l2_unused):
    beta_ = make2d(beta)
    if beta_.shape[1] != 1 and beta_.shape[0] == 1:
        beta_ = beta_.T
    res = objF_instance.compute(X, y, beta_)
    v = res.valueIdx[0, 0]
    v *= n
    return v


def _daal4py_grad_(beta, objF_instance, X, y, n, l2_unused):
    beta_ = make2d(beta)
    if beta_.shape[1] != 1 and beta_.shape[0] == 1:
        beta_ = beta_.T
    res = objF_instance.compute(X, y, beta_)
    # Copy is needed for newton-cg to work correctly
    # Otherwise evaluation at nearby point in line search
    # overwrites gradient at the starting point
    gr = res.gradientIdx.ravel()
    gr = gr * n  # force copy
    return gr


def _daal4py_grad_hess_(beta, objF_instance, X, y, n, l2):
    beta_ = make2d(beta)
    if beta_.shape[1] != 1 and beta_.shape[0] == 1:
        beta_ = beta_.T
    res = objF_instance.compute(X, y, beta_)
    gr = res.gradientIdx.ravel()
    gr = gr * n

    if isinstance(objF_instance, daal4py.optimization_solver_logistic_loss):
        # dealing with binary logistic regression
        # pp - array of probabilities for class=1, shape=(nSamples,)
        pp = beta_[0, 0] + np.dot(X, beta_[1:, 0])
        y2 = -1 + 2 * y[:, 0]
        pp *= y2
        np.exp(pp, out=pp)
        pp += 1
        np.reciprocal(pp, out=pp)
        np.square(pp, out=pp)
        del y2

        def hessp(v):
            pp0 = pp * (v[0] + np.dot(X, v[1:]))
            res = np.empty_like(v)
            res[0] = pp0.sum()
            res[1:] = np.dot(pp0, X)
            res[1:] += (2 * l2) * v[1:]
            return res
    else:
        # dealing with multi-class logistic regression
        beta__ = beta_.reshape((-1, 1 + X.shape[1]))  # (nClasses, nSamples)
        beta_shape = beta__.shape
        # pp - array of class probabilities, shape=(nSamples, nClasses)
        pp = beta__[np.newaxis, :, 0] + np.dot(X, beta__[:, 1:].T)
        pp -= pp.max(axis=1, keepdims=True)
        np.exp(pp, out=pp)
        pp /= pp.sum(axis=1, keepdims=True)

        def hessp(v):
            v2 = v.reshape(beta_shape)
            r_yhat = v2[np.newaxis, :, 0] + np.dot(X, v2[:, 1:].T)
            r_yhat += (-pp * r_yhat).sum(axis=1)[:, np.newaxis]
            r_yhat *= pp
            hessProd = np.zeros(beta_shape)
            hessProd[:, 1:] = np.dot(r_yhat.T, X) + (2 * l2) * v2[:, 1:]
            hessProd[:, 0] = r_yhat.sum(axis=0)
            return hessProd.ravel()

    return gr, hessp



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# daal4py PCA example for shared memory systems

import daal4py as d4p
import numpy as np

# let's try to use pandas' fast csv reader
try:
    import pandas

    def read_csv(f, c=None, t=np.float64):
        return pandas.read_csv(f, usecols=c, delimiter=',', header=None, dtype=t)
except ImportError:
    # fall back to numpy loadtxt
    def read_csv(f, c=None, t=np.float64):
        return np.loadtxt(f, usecols=c, delimiter=',', ndmin=2)


def main(readcsv=read_csv, method='svdDense'):
    infile = "./data/batch/pca_normalized.csv"

    # 'normalization' is an optional parameter to PCA;
    # we use z-score which could be configured differently
    zscore = d4p.normalization_zscore()
    # configure a PCA object
    algo = d4p.pca(resultsToCompute="mean|variance|eigenvalue",
                   isDeterministic=True, normalization=zscore)

    # let's provide a file directly, not a table/array
    result1 = algo.compute(infile)

    # We can also load the data ourselfs and provide the numpy array
    data = readcsv(infile)
    result2 = algo.compute(data)

    # PCA result objects provide eigenvalues, eigenvectors, means and variances
    assert np.allclose(result1.eigenvalues, result2.eigenvalues)
    assert np.allclose(result1.eigenvectors, result2.eigenvectors)
    assert np.allclose(result1.means, result2.means)
    assert np.allclose(result1.variances, result2.variances)
    assert result1.eigenvalues.shape == (1, data.shape[1])
    assert result1.eigenvectors.shape == (data.shape[1], data.shape[1])
    assert result1.means.shape == (1, data.shape[1])
    assert result1.variances.shape == (1, data.shape[1])

    return result1


if __name__ == "__main__":
    result1 = main()
    print("\nEigenvalues:\n", result1.eigenvalues)
    print("\nEigenvectors:\n", result1.eigenvectors)
    print("\nMeans:\n", result1.means)
    print("\nVariances:\n", result1.variances)
    print('All looks good!')



#===============================================================================
# Copyright 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from abc import ABCMeta

from numbers import Integral

import numpy as np
from ..datatypes import (
    _check_X_y,
    _check_array,
    _column_or_1d,
    _check_n_features,
    _check_classification_targets,
    _num_samples
)

from daal4py import (
    bf_knn_classification_training,
    bf_knn_classification_prediction,
    kdtree_knn_classification_training,
    kdtree_knn_classification_prediction
)
from onedal import _backend

from ..common._mixin import ClassifierMixin
from ..common._policy import _get_policy
from ..common._estimator_checks import _check_is_fitted, _is_classifier, _is_regressor
from ..datatypes._data_conversion import from_table, to_table


class NeighborsCommonBase(metaclass=ABCMeta):
    def _parse_auto_method(self, method, n_samples, n_features):
        result_method = method

        if (method in ['auto', 'ball_tree']):
            condition = self.n_neighbors is not None and \
                self.n_neighbors >= n_samples // 2
            if self.metric == 'precomputed' or n_features > 15 or condition:
                result_method = 'brute'
            else:
                if self.metric == 'euclidean':
                    result_method = 'kd_tree'
                else:
                    result_method = 'brute'

        return result_method

    def _validate_data(self, X, y=None, reset=True,
                       validate_separately=False, **check_params):
        if y is None:
            if self.requires_y:
                raise ValueError(
                    f"This {self.__class__.__name__} estimator "
                    f"requires y to be passed, but the target y is None."
                )
            X = _check_array(X, **check_params)
            out = X, y
        else:
            if validate_separately:
                # We need this because some estimators validate X and y
                # separately, and in general, separately calling _check_array()
                # on X and y isn't equivalent to just calling _check_X_y()
                # :(
                check_X_params, check_y_params = validate_separately
                X = _check_array(X, **check_X_params)
                y = _check_array(y, **check_y_params)
            else:
                X, y = _check_X_y(X, y, **check_params)
            out = X, y

        if check_params.get('ensure_2d', True):
            _check_n_features(self, X, reset=reset)

        return out

    def _get_weights(self, dist, weights):
        if weights in (None, "uniform"):
            return None
        if weights == "distance":
            # if user attempts to classify a point that was zero distance from one
            # or more training points, those training points are weighted as 1.0
            # and the other points as 0.0
            if dist.dtype is np.dtype(object):
                for point_dist_i, point_dist in enumerate(dist):
                    # check if point_dist is iterable
                    # (ex: RadiusNeighborClassifier.predict may set an element of
                    # dist to 1e-6 to represent an 'outlier')
                    if hasattr(point_dist, "__contains__") and 0.0 in point_dist:
                        dist[point_dist_i] = point_dist == 0.0
                    else:
                        dist[point_dist_i] = 1.0 / point_dist
            else:
                with np.errstate(divide="ignore"):
                    dist = 1.0 / dist
                inf_mask = np.isinf(dist)
                inf_row = np.any(inf_mask, axis=1)
                dist[inf_row] = inf_mask[inf_row]
            return dist
        elif callable(weights):
            return weights(dist)
        else:
            raise ValueError(
                "weights not recognized: should be 'uniform', "
                "'distance', or a callable function"
            )

    def _get_onedal_params(self, data):
        class_count = 0 if self.classes_ is None else len(self.classes_)
        weights = getattr(self, 'weights', 'uniform')
        return {
            'fptype': 'float' if data.dtype is np.dtype('float32') else 'double',
            'vote_weights': 'uniform' if weights == 'uniform' else 'distance',
            'method': self._fit_method,
            'radius': self.radius,
            'class_count': class_count,
            'neighbor_count': self.n_neighbors,
            'metric': self.effective_metric_,
            'p': self.p,
            'metric_params': self.effective_metric_params_,
            'result_option': 'indices|distances',
        }

    def _get_daal_params(self, data):
        class_count = 0 if self.classes_ is None else len(self.classes_)
        weights = getattr(self, 'weights', 'uniform')
        params = {
            'fptype': 'float' if data.dtype is np.dtype('float32') else 'double',
            'method': 'defaultDense',
            'k': self.n_neighbors,
            'voteWeights': 'voteUniform' if weights == 'uniform' else 'voteDistance',
            'resultsToCompute': 'computeIndicesOfNeighbors|computeDistances',
            'resultsToEvaluate': 'none' if getattr(self, '_y', None) is None
            else 'computeClassLabels'
        }
        if class_count != 0:
            params['nClasses'] = class_count
        return params


class NeighborsBase(NeighborsCommonBase, metaclass=ABCMeta):
    def __init__(self, n_neighbors=None, radius=None,
                 algorithm='auto', metric='minkowski',
                 p=2, metric_params=None):
        self.n_neighbors = n_neighbors
        self.radius = radius
        self.algorithm = algorithm
        self.metric = metric
        self.p = p
        self.metric_params = metric_params

    def _validate_targets(self, y, dtype):
        arr = _column_or_1d(y, warn=True)

        try:
            return arr.astype(dtype, copy=False)
        except ValueError:
            return arr

    def _validate_n_classes(self):
        if len(self.classes_) < 2:
            raise ValueError(
                "The number of classes has to be greater than one; got %d"
                " class" % len(self.classes_))

    def _fit(self, X, y, queue):
        self._onedal_model = None
        self._tree = None
        self._shape = None
        self.classes_ = None
        self.effective_metric_ = getattr(self, 'effective_metric_', self.metric)
        self.effective_metric_params_ = getattr(
            self, 'effective_metric_params_', self.metric_params)

        if y is not None or self.requires_y:
            X, y = super()._validate_data(X, y, dtype=[np.float64, np.float32])
            self._shape = y.shape

            if _is_classifier(self) or _is_regressor(self):
                if y.ndim == 1 or y.ndim == 2 and y.shape[1] == 1:
                    self.outputs_2d_ = False
                    y = y.reshape((-1, 1))
                else:
                    self.outputs_2d_ = True

                if _is_classifier(self):
                    _check_classification_targets(y)
                self.classes_ = []
                self._y = np.empty(y.shape, dtype=int)
                for k in range(self._y.shape[1]):
                    classes, self._y[:, k] = np.unique(
                        y[:, k], return_inverse=True)
                    self.classes_.append(classes)

                if not self.outputs_2d_:
                    self.classes_ = self.classes_[0]
                    self._y = self._y.ravel()
                if _is_classifier(self):
                    self._validate_n_classes()
            else:
                self._y = y
        else:
            X, _ = super()._validate_data(X, dtype=[np.float64, np.float32])

        self.n_samples_fit_ = X.shape[0]
        self.n_features_in_ = X.shape[1]
        self._fit_X = X

        if self.n_neighbors is not None:
            if self.n_neighbors <= 0:
                raise ValueError(
                    "Expected n_neighbors > 0. Got %d" %
                    self.n_neighbors
                )
            if not isinstance(self.n_neighbors, Integral):
                raise TypeError(
                    "n_neighbors does not take %s value, "
                    "enter integer value" %
                    type(self.n_neighbors))

        self._fit_method = super()._parse_auto_method(
            self.algorithm,
            self.n_samples_fit_, self.n_features_in_)

        if (_is_classifier(self) or _is_regressor(self)) and y.dtype != X.dtype:
            y = self._validate_targets(self._y, X.dtype).reshape((-1, 1))
        result = self._onedal_fit(X, y, queue)

        if y is not None and _is_regressor(self):
            self._y = y if self._shape is None else y.reshape(self._shape)

        self._onedal_model = result
        result = self

        return result

    def _kneighbors(self, X=None, n_neighbors=None,
                    return_distance=True, queue=None):
        n_features = getattr(self, 'n_features_in_', None)
        shape = getattr(X, 'shape', None)
        if n_features and shape and len(shape) > 1 and shape[1] != n_features:
            raise ValueError((f'X has {X.shape[1]} features, '
                              f'but kneighbors is expecting '
                              f'{n_features} features as input'))

        _check_is_fitted(self)

        if n_neighbors is None:
            n_neighbors = self.n_neighbors
        elif n_neighbors <= 0:
            raise ValueError(
                "Expected n_neighbors > 0. Got %d" %
                n_neighbors
            )
        else:
            if not isinstance(n_neighbors, Integral):
                raise TypeError(
                    "n_neighbors does not take %s value, "
                    "enter integer value" %
                    type(n_neighbors))

        if X is not None:
            query_is_train = False
            X = _check_array(X, accept_sparse='csr', dtype=[np.float64, np.float32])
        else:
            query_is_train = True
            X = self._fit_X
            # Include an extra neighbor to account for the sample itself being
            # returned, which is removed later
            n_neighbors += 1
        self.n_neighbors = n_neighbors

        n_samples_fit = self.n_samples_fit_
        if n_neighbors > n_samples_fit:
            raise ValueError(
                "Expected n_neighbors <= n_samples, "
                " but n_samples = %d, n_neighbors = %d" %
                (n_samples_fit, n_neighbors)
            )

        chunked_results = None
        method = super()._parse_auto_method(
            self._fit_method, self.n_samples_fit_, n_features)

        gpu_device = queue is not None and queue.sycl_device.is_gpu
        if self.effective_metric_ == 'euclidean' and not gpu_device:
            params = super()._get_daal_params(X)
        else:
            params = super()._get_onedal_params(X)

        prediction_results = self._onedal_predict(
            self._onedal_model, X, params, queue=queue)

        if self.effective_metric_ == 'euclidean' and not gpu_device:
            distances = prediction_results.distances
            indices = prediction_results.indices
        else:
            distances = from_table(prediction_results.distances)
            indices = from_table(prediction_results.indices)

        if method == 'kd_tree':
            for i in range(distances.shape[0]):
                seq = distances[i].argsort()
                indices[i] = indices[i][seq]
                distances[i] = distances[i][seq]

        if return_distance:
            results = distances, indices
        else:
            results = indices

        if chunked_results is not None:
            if return_distance:
                neigh_dist, neigh_ind = zip(*chunked_results)
                results = np.vstack(neigh_dist), np.vstack(neigh_ind)
            else:
                results = np.vstack(chunked_results)

        if not query_is_train:
            return results
        # If the query data is the same as the indexed data, we would like
        # to ignore the first nearest neighbor of every sample, i.e
        # the sample itself.
        distances = distances[:, 1:]
        indices = indices[:, 1:]

        if return_distance:
            neigh_dist, neigh_ind = results
        else:
            neigh_ind = results

        n_queries, _ = X.shape
        sample_range = np.arange(n_queries)[:, None]
        sample_mask = neigh_ind != sample_range

        # Corner case: When the number of duplicates are more
        # than the number of neighbors, the first NN will not
        # be the sample, but a duplicate.
        # In that case mask the first duplicate.
        dup_gr_nbrs = np.all(sample_mask, axis=1)
        sample_mask[:, 0][dup_gr_nbrs] = False

        neigh_ind = np.reshape(
            neigh_ind[sample_mask], (n_queries, n_neighbors - 1))

        if return_distance:
            neigh_dist = np.reshape(
                neigh_dist[sample_mask], (n_queries, n_neighbors - 1))
            return neigh_dist, neigh_ind
        return neigh_ind


class KNeighborsClassifier(NeighborsBase, ClassifierMixin):
    def __init__(self, n_neighbors=5, *,
                 weights='uniform', algorithm='auto',
                 p=2, metric='minkowski', metric_params=None, **kwargs):
        super().__init__(
            n_neighbors=n_neighbors,
            algorithm=algorithm,
            metric=metric, p=p,
            metric_params=metric_params,
            **kwargs)
        self.weights = weights

    def _get_onedal_params(self, data):
        params = super()._get_onedal_params(data)
        params['result_option'] = 'responses'
        return params

    def _get_daal_params(self, data):
        params = super()._get_daal_params(data)
        params['resultsToEvaluate'] = 'computeClassLabels'
        params['resultsToCompute'] = ''
        return params

    def _onedal_fit(self, X, y, queue):
        gpu_device = queue is not None and queue.sycl_device.is_gpu
        if self.effective_metric_ == 'euclidean' and not gpu_device:
            params = self._get_daal_params(X)
            if self._fit_method == 'brute':
                train_alg = bf_knn_classification_training

            else:
                train_alg = kdtree_knn_classification_training

            return train_alg(**params).compute(X, y).model

        policy = _get_policy(queue, X, y)
        params = self._get_onedal_params(X)
        train_alg = _backend.neighbors.classification.train(policy, params,
                                                            *to_table(X, y))

        return train_alg.model

    def _onedal_predict(self, model, X, params, queue):
        gpu_device = queue is not None and queue.sycl_device.is_gpu
        if self.effective_metric_ == 'euclidean' and not gpu_device:
            if self._fit_method == 'brute':
                predict_alg = bf_knn_classification_prediction

            else:
                predict_alg = kdtree_knn_classification_prediction

            return predict_alg(**params).compute(X, model)
        policy = _get_policy(queue, X)

        if hasattr(self, '_onedal_model'):
            model = self._onedal_model
        else:
            model = self._create_model(_backend.neighbors.classification)
        result = _backend.neighbors.classification.infer(
            policy, params, model, to_table(X))

        return result

    def fit(self, X, y, queue=None):
        return super()._fit(X, y, queue=queue)

    def predict(self, X, queue=None):
        X = _check_array(X, accept_sparse='csr', dtype=[np.float64, np.float32])
        onedal_model = getattr(self, '_onedal_model', None)
        n_features = getattr(self, 'n_features_in_', None)
        n_samples_fit_ = getattr(self, 'n_samples_fit_', None)
        shape = getattr(X, 'shape', None)
        if n_features and shape and len(shape) > 1 and shape[1] != n_features:
            raise ValueError((f'X has {X.shape[1]} features, '
                              f'but KNNClassifier is expecting '
                              f'{n_features} features as input'))

        _check_is_fitted(self)

        self._fit_method = super()._parse_auto_method(
            self.algorithm,
            n_samples_fit_, n_features)

        self._validate_n_classes()

        gpu_device = queue is not None and queue.sycl_device.is_gpu
        if self.effective_metric_ == 'euclidean' and not gpu_device:
            params = self._get_daal_params(X)
        else:
            params = self._get_onedal_params(X)

        prediction_result = self._onedal_predict(onedal_model, X, params, queue=queue)
        if self.effective_metric_ == 'euclidean' and not gpu_device:
            responses = prediction_result.prediction
        else:
            responses = from_table(prediction_result.responses)
        result = self.classes_.take(
            np.asarray(responses.ravel(), dtype=np.intp))

        return result

    def predict_proba(self, X, queue=None):
        neigh_dist, neigh_ind = self.kneighbors(X, queue=queue)

        classes_ = self.classes_
        _y = self._y
        if not self.outputs_2d_:
            _y = self._y.reshape((-1, 1))
            classes_ = [self.classes_]

        n_queries = _num_samples(X)

        weights = self._get_weights(neigh_dist, self.weights)
        if weights is None:
            weights = np.ones_like(neigh_ind)

        all_rows = np.arange(n_queries)
        probabilities = []
        for k, classes_k in enumerate(classes_):
            pred_labels = _y[:, k][neigh_ind]
            proba_k = np.zeros((n_queries, classes_k.size))

            # a simple ':' index doesn't work right
            for i, idx in enumerate(pred_labels.T):  # loop is O(n_neighbors)
                proba_k[all_rows, idx] += weights[:, i]

            # normalize 'votes' into real [0,1] probabilities
            normalizer = proba_k.sum(axis=1)[:, np.newaxis]
            normalizer[normalizer == 0.0] = 1.0
            proba_k /= normalizer

            probabilities.append(proba_k)

        if not self.outputs_2d_:
            probabilities = probabilities[0]

        return probabilities

    def kneighbors(self, X=None, n_neighbors=None,
                   return_distance=True, queue=None):
        return super()._kneighbors(X, n_neighbors, return_distance, queue=queue)


class NearestNeighbors(NeighborsBase):
    def __init__(self, n_neighbors=5, *,
                 weights='uniform', algorithm='auto',
                 p=2, metric='minkowski', metric_params=None, **kwargs):
        super().__init__(
            n_neighbors=n_neighbors,
            algorithm=algorithm,
            metric=metric, p=p,
            metric_params=metric_params,
            **kwargs)
        self.weights = weights

    def _get_onedal_params(self, data):
        params = super()._get_onedal_params(data)
        params['result_option'] = 'indices|distances'
        return params

    def _get_daal_params(self, data):
        params = super()._get_daal_params(data)
        params['resultsToCompute'] = 'computeIndicesOfNeighbors|computeDistances'
        params['resultsToEvaluate'] = 'none' if getattr(self, '_y', None) is None \
            else 'computeClassLabels'
        return params

    def _onedal_fit(self, X, y, queue):
        gpu_device = queue is not None and queue.sycl_device.is_gpu
        if self.effective_metric_ == 'euclidean' and not gpu_device:
            params = self._get_daal_params(X)
            if self._fit_method == 'brute':
                train_alg = bf_knn_classification_training

            else:
                train_alg = kdtree_knn_classification_training

            return train_alg(**params).compute(X, y).model

        policy = _get_policy(queue, X, y)
        params = self._get_onedal_params(X)
        train_alg = _backend.neighbors.search.train(policy, params,
                                                    to_table(X))

        return train_alg.model

    def _onedal_predict(self, model, X, params, queue):
        gpu_device = queue is not None and queue.sycl_device.is_gpu
        if self.effective_metric_ == 'euclidean' and not gpu_device:
            if self._fit_method == 'brute':
                predict_alg = bf_knn_classification_prediction

            else:
                predict_alg = kdtree_knn_classification_prediction

            return predict_alg(**params).compute(X, model)

        policy = _get_policy(queue, X)

        if hasattr(self, '_onedal_model'):
            model = self._onedal_model
        else:
            model = self._create_model(_backend.neighbors.search)
        result = _backend.neighbors.search.infer(policy, params, model, to_table(X))

        return result

    def fit(self, X, y, queue=None):
        return super()._fit(X, y, queue=queue)

    def kneighbors(self, X=None, n_neighbors=None,
                   return_distance=True, queue=None):
        return super()._kneighbors(X, n_neighbors, return_distance, queue=queue)
